from aiogram.types import User, InlineKeyboardMarkup, InlineKeyboardButton, Message
from NekoGram import Neko, Menu, NekoRouter
from typing import Dict, List
try:
    import ujson as json
except ImportError:
    import json

ROUTER: NekoRouter = NekoRouter(name='broadcast')


@ROUTER.formatter()
async def widget_broadcast(data: Menu, user: User, neko: Neko):
    if not await neko.storage.check('SELECT id FROM nekogram_users WHERE id != %s', user.id):
        await data.obj.answer(text=data.extras['alt_text'], show_alert=True)
        data.break_execution()


@ROUTER.formatter()
async def widget_broadcast_post_markup(data: Menu, user: User, neko: Neko):
    user_data = await neko.storage.get_user_data(user_id=user.id)
    raw_markup: List[List[Dict[str, str]]] = user_data.get('widget_broadcast_post_markup', [])
    markup: InlineKeyboardMarkup = InlineKeyboardMarkup()
    row_index: int = 0

    for row in raw_markup:
        row_buttons = []

        button_index: int = 0
        for button in row:
            row_buttons.append(InlineKeyboardButton(text=button['text'], callback_data=f'widget_broadcast_remove_button'
                                                                                       f'#{row_index}-{button_index}'))
            button_index += 1
        if button_index < 3:
            row_buttons.append(InlineKeyboardButton(text='➕',
                                                    callback_data=f'widget_broadcast_add_button_step_1#{row_index}'))
        markup.add(*row_buttons)
        row_index += 1

    await data.build(markup=markup, markup_format={'row_index': row_index})


@ROUTER.formatter()
async def widget_broadcast_add_button_step_1(data: Menu, user: User, neko: Neko):
    if isinstance(data.obj, Message):
        return
    await neko.storage.set_user_data(data={'widget_broadcast_post_row_index': data.call_data}, user_id=user.id)



from aiogram import exceptions as aiogram_exceptions, types
from typing import List, Optional, Union
from contextlib import suppress
import NekoGram


def _convert_call_data(call_data: List[str]) -> Optional[Union[str, int]]:
    if len(call_data) == 1:
        return None

    call_data = call_data[1]
    return int(call_data) if call_data.isnumeric() else call_data


async def menu_callback_query_handler(call: types.CallbackQuery):
    neko: NekoGram.Neko = call.conf['neko']
    if call.data == 'menu_start':  # Reset user data on start
        await neko.storage.set_user_data(user_id=call.from_user.id)

    call_data = call.data.split(neko.callback_parameters_delimiter)
    current_menu = await neko.build_menu(name=call_data[0], obj=call, callback_data=_convert_call_data(call_data))
    if current_menu is None:
        return

    await neko.storage.set_user_menu(user_id=call.from_user.id, menu=current_menu.name)
    if not current_menu.filters and neko.functions.get(current_menu.name):  # Call function if it doesn't need input
        await neko.functions[current_menu.name](current_menu, call, neko)
        return

    try:
        await current_menu.edit_text()
    except aiogram_exceptions.InlineKeyboardExpected:
        with suppress(Exception):
            await call.message.delete()
        await current_menu.send_message()
    except (aiogram_exceptions.MessageCantBeEdited, aiogram_exceptions.MessageToEditNotFound):
        await current_menu.send_message()



from NekoGram import NekoRouter, Menu, Neko
from aiogram import types

ROUTER: NekoRouter = NekoRouter(name='languages')


@ROUTER.function()
async def widget_languages_set(data: Menu, call: types.CallbackQuery, neko: Neko):
    await neko.storage.set_user_language(language=data.call_data, user_id=call.from_user.id)
    data = await neko.build_menu(name=data.name, obj=call)
    await call.answer(text=data.text)
    data = await neko.build_menu(name=data.next_menu, obj=call)
    await data.build()
    await data.edit_text()



from ...base_neko import BaseNeko
from typing import Dict, List
try:
    import ujson as json
except ImportError:
    import json

replacements: Dict[str, str]
languages: List[str]


async def startup(neko: BaseNeko):
    global replacements
    global languages
    with open('NekoGram/widgets/languages/replacements.json', 'r') as f:
        replacements = json.load(f)
    languages = list(neko.text_processor.texts.keys())



from typing import Union, Optional, Any, TextIO, Dict
from .base_processor import BaseProcessor
from os.path import isdir, isfile
from io import TextIOWrapper
from os import listdir

try:
    import ujson as json
except ImportError:
    import json


class JSONProcessor(BaseProcessor):
    def __init__(self):
        super().__init__()

    def _verify(self):
        for lang, data in self.texts.items():
            if data.get('start') is None:
                raise RuntimeError(f'\"start\" menu is undefined for {lang}! *Nervous paw shaking*')

    def _add_text(self, text_data: dict):
        for lang, text in text_data.items():
            if lang not in self.texts:
                self.texts[lang] = text
            else:
                self.texts[lang].update(text)

    def add_texts(self, texts: Union[Dict[str, Any], TextIO, str] = 'translations',
                  is_widget: bool = False) -> Dict[str, Dict[str, Any]]:
        """
        Assigns a required piece of texts to use later
        :param texts: Dictionary or JSON containing texts, path to a file or path to a directory containing texts
        :param is_widget: Set true if text path is for a widget
        :return: Processed texts
        """
        processed_texts: Dict[str, Dict[str, Any]] = dict()
        processed_json: Optional[Dict[str, Any]] = None

        if isinstance(texts, str):
            if isdir(texts):  # Path to directory containing translations
                text_list = listdir(texts)
                for file in text_list:
                    if file.endswith('.json'):
                        with open(f'{texts}/{file}', 'r', encoding='utf-8') as text_file:
                            processed_json = json.load(text_file)
                            lang: Optional[str] = processed_json.get('lang')
                            if lang is None:
                                raise ValueError(f'The supplied translation file does not contain a language '
                                                 f'definition ("lang" field)')
                            processed_texts[lang] = processed_json
                self._add_text(text_data=processed_texts)
                if not is_widget:
                    self._verify()
                return processed_texts

            elif not isfile(texts):  # String JSON
                processed_json = json.loads(texts)
            elif isfile(texts):  # File path
                with open(texts, 'r', encoding='utf-8') as file:
                    processed_json = json.load(file)

        elif isinstance(texts, TextIOWrapper):  # IO JSON file
            processed_json = json.load(texts)
        else:
            raise ValueError('No valid text path, text or language supplied')
        if processed_json:
            lang: Optional[str] = processed_json.get('lang')
            if lang is None:
                raise ValueError(f'The supplied translation file does not contain a language definition ("lang" field)')
            processed_texts[lang] = processed_json

        self._add_text(text_data=processed_texts)
        if not is_widget:
            self._verify()
        return processed_texts

    @property
    def processor_type(self) -> str:
        return 'json'



from typing import Union, Optional, Dict, Any, AsyncGenerator, List
from datetime import datetime, timedelta
from abc import ABC, abstractmethod


class BaseStorage(ABC):
    def __init__(self, default_language: str = 'en'):
        self.users: Dict[str, Dict[str, Any]] = dict()
        self.default_language: str = default_language
        self._cached_user_languages: Dict[str, Dict[str, Union[str, datetime]]] = dict()

    @abstractmethod
    async def set_user_language(self, user_id: int, language: str):
        self._cached_user_languages[str(user_id)] = {'date': datetime.now(), 'lang': language}

    @abstractmethod
    async def get_user_language(self, user_id: int) -> str:
        pass

    async def get_cached_user_language(self, user_id: Union[int, str]) -> Optional[str]:
        user_id = str(user_id)
        if (user_id in self._cached_user_languages.keys() and
                self._cached_user_languages[user_id]['date'] + timedelta(minutes=20) > datetime.now()):
            return self._cached_user_languages[user_id]['lang']
        else:
            return None

    @abstractmethod
    async def set_user_data(self, user_id: int, data: Optional[Dict[str, Any]] = None,
                            replace: bool = False) -> Dict[str, Any]:
        pass

    @abstractmethod
    async def get_user_data(self, user_id: int) -> Dict[str, Any]:
        pass

    @abstractmethod
    async def set_user_menu(self, user_id: int, menu: Optional[str] = None):
        pass

    @abstractmethod
    async def get_user_menu(self, user_id: int) -> Optional[str]:
        pass

    @abstractmethod
    async def check_user_exists(self, user_id: int) -> bool:
        pass

    @abstractmethod
    async def set_last_message_id(self, user_id: int, message_id: int):
        pass

    @abstractmethod
    async def get_last_message_id(self, user_id: int) -> Optional[int]:
        pass

    @abstractmethod
    async def create_user(self, user_id: int, language: Optional[str] = None):
        pass

    @abstractmethod
    async def apply(self, query, args=None):
        pass

    async def select(self, query, args=None) -> AsyncGenerator[Dict[str, Any], None]:
        yield

    @abstractmethod
    async def get(self, query: str, args=None, fetch_all: bool = False) -> Union[bool, List[Dict[str, Any]],
                                                                                 Dict[str, Any]]:
        pass

    @abstractmethod
    async def check(self, query: str, args=None) -> int:
        pass

    async def acquire_pool(self):
        pass



from aiogram.types import InlineKeyboardMarkup, InlineKeyboardButton
from typing import List, Dict
from ...base_neko import BaseNeko


def assemble_markup(markup: List[List[Dict[str, str]]]) -> InlineKeyboardMarkup:
    final_markup = InlineKeyboardMarkup()
    for row in markup:
        row_buttons = []
        for button in row:
            row_buttons.append(InlineKeyboardButton(text=button['text'], url=button['url']))
        final_markup.add(*row_buttons)

    return final_markup


async def send_post(user_data: dict, chat_id: int, neko: BaseNeko):
    post_data = user_data['widget_broadcast']
    post_data['content_type'] = user_data['widget_broadcast_content_type']
    post_data['caption'] = user_data.get('widget_broadcast_post_caption')
    raw_markup = user_data.get('widget_broadcast_post_markup', [])

    if post_data['content_type'] == 'text':
        await neko.bot.send_message(text=post_data['caption'], parse_mode='HTML', disable_web_page_preview=True,
                                    reply_markup=assemble_markup(raw_markup), chat_id=chat_id)
    elif post_data['content_type'] == 'photo':
        await neko.bot.send_photo(photo=post_data['file_id'], parse_mode='HTML', caption=post_data['caption'],
                                  reply_markup=assemble_markup(raw_markup), chat_id=chat_id)
    elif post_data['content_type'] == 'video':
        await neko.bot.send_video(video=post_data['file_id'], parse_mode='HTML', caption=post_data['caption'],
                                  reply_markup=assemble_markup(raw_markup), chat_id=chat_id)
    else:
        await neko.bot.send_animation(animation=post_data['file_id'], parse_mode='HTML', caption=post_data['caption'],
                                      reply_markup=assemble_markup(raw_markup), chat_id=chat_id)



from typing import Optional, Union, Any, AsyncGenerator, List, Dict, Tuple
from pymysql import err as mysql_errors
from ..base_storage import BaseStorage
from ...logger import LOGGER
from pymysql.constants import CLIENT
from contextlib import suppress
import importlib_resources
import aiomysql

try:
    from aiomysql.cursors import DictCursor
except ImportError:
    raise ImportError('Install aiomysql to use MySQLStorage!')

try:
    import ujson as json
except ImportError:
    import json


class MySQLStorage(BaseStorage):
    def __init__(self, database: str, host: str = 'localhost', port: int = 3306, user: str = 'root',
                 password: Optional[str] = None, default_language: str = 'en'):
        """
        Initialize database
        :param database: Database name
        :param host: Database host
        :param port: Database port
        :param user: Database user
        :param password: Database password
        """

        self.pool: Optional[aiomysql.Pool] = None
        self.host: str = host
        self.port: int = port
        self.user: str = user
        self.password: str = password
        self.database = database

        file = importlib_resources.files('NekoGram.storages.mysql').joinpath('tables.json').read_text(encoding='utf-8')
        self._table_structs: Dict[str, Dict[str, Dict[str, Optional[str]]]] = json.loads(file)
        super().__init__(default_language=default_language)

    def __del__(self):
        self.pool.close()

    async def verify_table(self, table: str, required_by: str):
        r = await self.get(f'DESCRIBE {table}', fetch_all=True)
        structure = self._table_structs[table]
        if isinstance(r, list):  # Table exists
            r: Dict[str, Dict[str, Optional[str]]] = {x['Field']: x for x in r}
            table_struct: Dict[str, Dict[str, Optional[str]]] = {x['Field']: x for x in structure.values()}
            for key, value in table_struct.items():
                sql_code = value['struct']
                value.pop('struct')
                if not r.get(key):  # Field does not exist
                    await self.apply(f'ALTER TABLE {table} ADD {sql_code}')
                    LOGGER.warning(f'Added {key} to {table} required by {required_by}.')
                elif r[key] != value:  # Field is defined in a wrong way
                    await self.apply(f'ALTER TABLE {table} MODIFY {sql_code}')
                    LOGGER.warning(f'Altered {key} in {table} required by {required_by}.')

        else:  # Table does not exist
            LOGGER.warning(f'Table {table} required by {required_by} does not exist, creating..')
            await self.apply(f'CREATE TABLE {table} ({",".join([f["struct"] for f in structure.values()])}) '
                             f'ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_unicode_ci;')

    async def add_tables(self, structure: Dict[str, Dict[str, Dict[str, Optional[str]]]], required_by: str):
        self._table_structs.update(structure)
        for table in structure.keys():
            await self.verify_table(table=table, required_by=required_by)

    async def _verify(self):
        connection = await aiomysql.connect(host=self.host, user=self.user, password=self.password, port=self.port,
                                            client_flag=CLIENT.MULTI_STATEMENTS)
        async with connection.cursor(DictCursor) as cursor:
            await cursor.execute('SHOW DATABASES LIKE %s', self.database)
            if not await cursor.fetchone():
                LOGGER.warning(f'MySQL database {self.database} does not exist, creating..')
                await cursor.execute(f'CREATE DATABASE IF NOT EXISTS {self.database} DEFAULT CHARACTER SET '
                                     f'utf8mb4 COLLATE utf8mb4_unicode_ci')
                await connection.commit()
        connection.close()

    async def acquire_pool(self):
        """
        Creates a new MySQL pool
        """
        LOGGER.info('Verifying database existence..')
        await self._verify()
        if isinstance(self.pool, aiomysql.Pool):
            with suppress(Exception):
                self.pool.close()

        self.pool = await aiomysql.create_pool(host=self.host, port=self.port, user=self.user,
                                               password=self.password, db=self.database)
        LOGGER.info('Verifying table structures, hold tight..')
        await self.verify_table(table='nekogram_users', required_by='NekoGram')
        LOGGER.info('MySQLStorage initialized successfully. ~nya')

    @staticmethod
    def _verify_args(args: Optional[Union[Tuple[Union[Any, Dict[str, Any]], ...], Any]]):
        if args is None:
            args = tuple()
        if not isinstance(args, (tuple, dict)):
            args = (args,)
        return args

    async def apply(self, query: str, args: Optional[Union[Tuple[Any, ...], Dict[str, Any], Any]] = None) -> int:
        """
        Executes SQL query and returns the number of affected rows
        :param query: SQL query to execute
        :param args: Arguments passed to the SQL query
        :return: Number of affected rows
        """
        args = self._verify_args(args)
        async with self.pool.acquire() as conn:
            async with conn.cursor(DictCursor) as cursor:
                try:
                    await cursor.execute(query, args)
                    await conn.commit()
                except mysql_errors.Error as e:
                    print(e)
                    await conn.rollback()

                if 'insert into' in query.lower():
                    return cursor.lastrowid
                else:
                    return cursor.rowcount

    async def select(self, query: str, args: Optional[Union[Tuple[Any, ...], Dict[str, Any], Any]] = None) -> \
            AsyncGenerator[Dict[str, Any], None]:
        """
        Generator that yields rows
        :param query: SQL query to execute
        :param args: Arguments passed to the SQL query
        :return: Yields rows one by one
        """
        args = self._verify_args(args)
        async with self.pool.acquire() as conn:
            async with conn.cursor(DictCursor) as cursor:
                try:
                    await cursor.execute(query, args)
                    await conn.commit()
                    while True:
                        item = await cursor.fetchone()
                        if item:
                            yield item
                        else:
                            break
                except mysql_errors.Error:
                    pass

    async def get(self, query: str, args: Optional[Union[Tuple[Any, ...], Dict[str, Any], Any]] = None,
                  fetch_all: bool = False) -> Union[bool, List[Dict[str, Any]], Dict[str, Any]]:
        """
        Get a single row or a list of rows from the database
        :param query: SQL query to execute
        :param args: Arguments passed to the SQL query
        :param fetch_all: Set True if you need a list of rows instead of just a single row
        :return: A row or a list or rows
        """
        args = self._verify_args(args)
        async with self.pool.acquire() as conn:
            async with conn.cursor(DictCursor) as cursor:
                try:
                    await cursor.execute(query, args)
                    await conn.commit()

                    if fetch_all:
                        return await cursor.fetchall()
                    else:
                        result = await cursor.fetchone()
                        return result if result else dict()
                except mysql_errors.Error:
                    return False

    async def check(self, query: str, args: Optional[Union[Tuple[Any, ...], Dict[str, Any], Any]] = None) -> int:
        args = self._verify_args(args)
        async with self.pool.acquire() as conn:
            async with conn.cursor(DictCursor) as cursor:
                try:
                    await cursor.execute(query, args)
                    await conn.commit()

                    return cursor.rowcount
                except mysql_errors.Error:
                    return 0

    async def set_user_language(self, user_id: int, language: str):
        await super().set_user_language(user_id=user_id, language=language)
        await self.apply('UPDATE nekogram_users SET lang = %s WHERE id = %s', (language, user_id))

    async def get_user_language(self, user_id: int) -> str:
        """
        Get user's language
        :param user_id: Telegram ID of the user
        :return: User's language
        """
        lang = await self.get_cached_user_language(user_id=user_id)
        if lang is None:
            lang = (await self.get('SELECT lang FROM nekogram_users WHERE id=%s',
                                   user_id)).get('lang', self.default_language)
        await super().set_user_language(user_id=user_id, language=lang)
        return lang

    async def get_user_data(self, user_id: int) -> Union[Dict[str, Any], bool]:
        """
        Get user data
        :param user_id: Telegram ID of the user
        :return: Decoded JSON user data
        """
        try:
            return json.loads((await self.get('SELECT data FROM nekogram_users WHERE id=%s',
                                              user_id)).get('data', '{}'))
        except TypeError:
            return False

    async def set_user_data(self, user_id: int, data: Optional[Dict[str, Any]] = None,
                            replace: bool = False) -> Dict[str, Any]:
        if data is None:
            data = dict()
            replace = True

        if replace:
            user_data = data
        else:
            user_data = await self.get_user_data(user_id=user_id)
            user_data.update(data)

        await self.apply('UPDATE nekogram_users SET data = %s WHERE id = %s', (json.dumps(user_data), user_id))
        return user_data

    async def set_user_menu(self, user_id: int, menu: Optional[str] = None):
        await self.set_user_data(user_id=user_id, data={'menu': menu})
        return menu

    async def get_user_menu(self, user_id: int) -> Optional[str]:
        return (await self.get_user_data(user_id=user_id)).get('menu')

    async def check_user_exists(self, user_id: int) -> bool:
        return bool(await self.check('SELECT id FROM nekogram_users WHERE id = %s', user_id))

    async def set_last_message_id(self, user_id: int, message_id: int):
        await self.apply('UPDATE nekogram_users SET last_message_id = %s WHERE id = %s', (message_id, user_id))

    async def get_last_message_id(self, user_id: int) -> Optional[int]:
        return (await self.get('SELECT last_message_id FROM nekogram_users WHERE id = %s',
                               user_id)).get('last_message_id')

    async def create_user(self, user_id: int, language: Optional[str] = None):
        if language is None:
            language = self.default_language

        await self.apply('INSERT INTO nekogram_users (id, lang) VALUES (%s, %s)', (user_id, language))



from typing import Optional, Union, Dict, List, Any, Type
from contextlib import suppress
from .base_neko import BaseNeko
from aiogram import types
from .logger import LOGGER


class Menu:
    _default_keyboard_values: Dict[str, str] = {'callback_data': 'call_data', 'switch_inline_query': 'query',
                                                'switch_inline_query_current_chat': 'cc_query',
                                                'text': 'text', 'url': 'url'}
    _inline_markup_identifiers: List[str] = ['call_data', 'callback_data', 'query', 'switch_inline_query', 'cc_query',
                                             'switch_inline_query_current_chat', 'url']

    def __init__(self, name: str, obj: Union[types.Message, types.InlineQuery],
                 markup: Optional[List[List[Dict[str, str]]]] = None, markup_row_width: Optional[int] = None,
                 text: Optional[str] = None, no_preview: Optional[bool] = None, parse_mode: Optional[str] = None,
                 silent: Optional[bool] = None, validation_error: Optional[str] = None,
                 extras: Optional[Dict[str, Any]] = None, keyboard_values_to_format: Optional[List[str]] = None,
                 markup_type: Optional[str] = None, prev_menu: Optional[str] = None, next_menu: Optional[str] = None,
                 filters: Optional[List[str]] = None, callback_data: Optional[Union[str, int]] = None):
        self.name: str = name
        self.obj: Optional[Union[types.Message, types.CallbackQuery]] = obj
        self._neko: BaseNeko = obj.conf['neko']
        self.markup: Optional[Union[types.InlineKeyboardMarkup, types.ReplyKeyboardMarkup]] = None

        self.text: Optional[str] = text
        self.no_preview: Optional[bool] = no_preview
        self.parse_mode: Optional[str] = parse_mode
        self.silent: Optional[bool] = silent
        self.raw_markup: Optional[List[List[Dict[str, str]]]] = markup
        self.markup_row_width: Optional[int] = markup_row_width
        self.validation_error: str = validation_error or 'ValidationError'
        self.extras: Dict[str, Any] = extras or dict()
        self.keyboard_values_to_format = keyboard_values_to_format or list(self._default_keyboard_values.keys())
        self.markup_type: Optional[str] = markup_type
        self.prev_menu: Optional[str] = prev_menu
        self.next_menu: Optional[str] = next_menu
        self.filters: Optional[List[str]] = filters
        self._call_data: Optional[Union[str, int]] = callback_data
        self._break_execution: bool = False

    def break_execution(self):
        self._break_execution = True

    @property
    def is_broken(self) -> bool:
        return self._break_execution

    @staticmethod
    def _apply_formatting(formatting: Optional[Union[List[Any], Dict[str, Any], Any]] = None, *items) -> List[str]:
        result = list()
        for item in items:
            if item is None:
                result.append(item)
                continue

            if isinstance(formatting, list):
                result.append(item.format(*formatting))
            elif isinstance(formatting, dict):
                result.append(item.format(**formatting))
            elif formatting is not None:
                result.append(item.format(formatting))
            else:
                result.append(item)
        return result

    async def send_message(self, user_id: Optional[int] = None) -> types.Message:
        if user_id is None:
            user_id = self.obj.from_user.id
        msg = await self.obj.bot.send_message(chat_id=user_id, text=self.text,
                                              parse_mode=self.parse_mode, disable_web_page_preview=self.no_preview,
                                              disable_notification=self.silent, reply_markup=self.markup)
        last_message_id = await self._neko.storage.get_last_message_id(user_id=user_id)
        await self._neko.storage.set_last_message_id(user_id=user_id, message_id=msg.message_id)
        with suppress(Exception):
            await self.obj.bot.delete_message(chat_id=user_id, message_id=last_message_id)
        return msg

    async def edit_text(self) -> types.Message:
        obj = self.obj if isinstance(self.obj, types.Message) else self.obj.message
        msg = await obj.edit_text(text=self.text, parse_mode=self.parse_mode, reply_markup=self.markup,
                                  disable_web_page_preview=self.no_preview)
        if isinstance(self.obj, types.CallbackQuery):
            await self._neko.storage.set_last_message_id(user_id=self.obj.from_user.id, message_id=msg.message_id)
        return msg

    async def _resolve_markup_type(self) -> Type[Union[types.InlineKeyboardMarkup, types.ReplyKeyboardMarkup]]:
        """
        Resolve markup type
        :return: InlineKeyboardMarkup or ReplyKeyboardMarkup type
        """
        if self.markup_type == 'inline':
            return types.InlineKeyboardMarkup
        elif self.markup_type == 'reply':
            return types.ReplyKeyboardMarkup

        for row in self.raw_markup:  # Guess the type otherwise
            for button in row:
                if any([i in button for i in self._inline_markup_identifiers]):
                    return types.InlineKeyboardMarkup
        return types.ReplyKeyboardMarkup

    async def _format_markup(self, markup: Union[types.InlineKeyboardMarkup, types.ReplyKeyboardMarkup],
                             markup_format: Optional[Union[List[Any], Dict[str, Any], Any]],
                             allowed_buttons: Optional[List[Union[str, int]]]) -> \
            Union[types.InlineKeyboardMarkup, types.ReplyKeyboardMarkup]:
        """
        Apply markup format
        :param markup: Base markup object
        :param markup_format: A list, dict or single value to use for formatting
        :param allowed_buttons: A list of allowed button IDs
        :return: A formatted markup object
        """
        filter_buttons: bool = isinstance(allowed_buttons, list)

        button_type = types.InlineKeyboardButton \
            if isinstance(markup, types.InlineKeyboardMarkup) else types.KeyboardButton

        for row in self.raw_markup:  # Fill existing markup
            buttons: Union[List[types.InlineKeyboardButton], List[types.KeyboardButton]] = list()
            for button in row:
                if filter_buttons and button.get('id') is not None and button['id'] not in allowed_buttons:
                    continue  # Button should be ignored

                if button_type == types.InlineKeyboardButton:  # Inline buttons only
                    for key, value in {'call_data': 'callback_data', 'query': 'switch_inline_query',
                                       'cc_query': 'switch_inline_query_current_chat'}.items():  # Apply aliases
                        if button.get(key) is not None:
                            button[value] = button.pop(key)

                    if button.get('url') and button['url'].startswith('@'):
                        button['url'] = button['url'].replace('@', 'https://t.me/')

                if markup_format:  # Apply button formatting
                    for item in self.keyboard_values_to_format:
                        if button.get(item):
                            button[item] = self._apply_formatting(markup_format, button[item])[0]
                buttons.append(button_type(**button))
            markup.add(*buttons)
        return markup

    async def build(self, text_format: Optional[Union[List[Any], Dict[str, Any], Any]] = None,
                    markup_format: Optional[Union[List[Any], Dict[str, Any], Any]] = None,
                    markup: Optional[Union[types.InlineKeyboardMarkup, types.ReplyKeyboardMarkup]] = None,
                    allowed_buttons: List[Union[str, int]] = None):
        """
        Build a menu
        :param text_format: Formatting for menu text
        :param markup_format: Formatting for markup buttons
        :param markup: Aiogram markup object
        :param allowed_buttons: In case of access limitation which buttons to display
        """
        if self.text:
            self.text = self._apply_formatting(text_format, self.text)[0]

        if markup is None and self.raw_markup is not None:
            # Resolve markup type
            markup_type = await self._resolve_markup_type()
            if markup_type == types.ReplyKeyboardMarkup:
                markup = types.ReplyKeyboardMarkup(row_width=self.markup_row_width or 3, resize_keyboard=True)
            else:
                markup = markup_type(row_width=self.markup_row_width or 3)

        if markup:
            self.markup = await self._format_markup(markup=markup, markup_format=markup_format,
                                                    allowed_buttons=allowed_buttons)

    @property
    def call_data(self) -> Optional[Any]:
        return self._call_data

    async def add_pagination(self, offset: int, found: int, limit: int, shift_last: bool = True):
        """
        Add a pagination
        :param offset: Item offset
        :param found: Number of found items found on this page
        :param limit: Max number of items that can be displayed at a time
        :param shift_last: Whether to shift the last button
        """
        if self.markup:
            LOGGER.warning(f'Pagination was not applied for {self.name} since the menu is already built!')
        last_button = self.raw_markup[-1]
        if shift_last:
            del self.raw_markup[-1]
        if offset >= limit and found > limit:
            # Add previous and next buttons
            self.raw_markup.append([{'call_data': f'{self.name}#{offset - limit}', 'text': '⬅️'},
                                    {'call_data': f'{self.name}#{offset + limit}', 'text': '➡️'}])
        elif offset >= limit:
            self.raw_markup.append([{'call_data': f'{self.name}#{offset - limit}', 'text': '⬅️'}])
        elif found > limit:
            self.raw_markup.append([{'call_data': f'{self.name}#{offset + limit}', 'text': '➡️'}])
        if shift_last:
            self.raw_markup.append(last_button)



from NekoGram import Neko, Menu, NekoRouter
from . import util
from aiogram import types

ROUTER: NekoRouter = NekoRouter(name='languages')


@ROUTER.formatter()
async def widget_languages(data: Menu, _: types.User, __: Neko):
    markup: types.InlineKeyboardMarkup = types.InlineKeyboardMarkup()
    for lang in util.languages:
        text = util.replacements.get(lang, lang).upper()
        text = f'{chr(ord(text[0]) + 127397) + chr(ord(text[1]) + 127397)}{lang.upper()}'
        markup.add(types.InlineKeyboardButton(text=text, callback_data=f'widget_languages_set#{lang}'))
    await data.build(markup=markup)



from aiogram.dispatcher.middlewares import BaseMiddleware
from .base_neko import BaseNeko
from aiogram import types
from typing import Union
from io import BytesIO
import aiohttp
try:
    import ujson as json
except ImportError:
    import json


class HandlerInjector(BaseMiddleware):
    """
    Neko injector middleware
    """

    def __init__(self, neko: BaseNeko):
        super().__init__()
        self.neko: BaseNeko = neko

    async def on_process_message(self, message: types.Message, _: dict):
        """
        This handler is called when dispatcher receives a message
        """
        # Get current handler
        message.conf['neko'] = self.neko

    async def on_process_callback_query(self, call: types.CallbackQuery, _: dict):
        """
        This handler is called when dispatcher receives a call
        """
        # Get current handler
        call.conf['neko'] = self.neko
        call.message.conf['neko'] = self.neko
        call.message.from_user = call.from_user


async def telegraph_upload(f: BytesIO, mime: str = 'image/png') -> Union[str, bool]:
    """
    Upload a file to Telegra.ph
    :param f: File BytesIO
    :param mime: File MIME type
    :return: File URL on success
    """
    # f = await (max(message.photo, key=lambda c: c.width)).download(destination=BytesIO())
    data = aiohttp.FormData()
    data.add_field('file', f.read(), filename=f'file.{mime.split("/")[1]}', content_type=mime)
    try:
        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=False),
                                         json_serialize=json.dumps) as session:
            async with session.post(url='https://telegra.ph/upload', data=data) as r:
                r = await r.json()
                if isinstance(r, dict) and r.get('error'):
                    return False
                if not r[-1]["src"].startswith('/'):
                    r[-1]["src"] = '/' + r[-1]["src"]
                return f'https://telegra.ph{r[-1]["src"]}'
    except aiohttp.ClientError:
        return False



from aiogram import exceptions as aiogram_exc, types
from NekoGram import Neko, Menu, NekoRouter
from NekoGram.utils import telegraph_upload
from typing import Union, Dict, List
from asyncio import sleep
from io import BytesIO
from . import utils

ROUTER: NekoRouter = NekoRouter(name='broadcast')


@ROUTER.function()
async def widget_broadcast(_: Menu, message: Union[types.Message, types.CallbackQuery], neko: Neko):
    user_data = await neko.storage.get_user_data(user_id=message.from_user.id)
    user_data.pop('menu')
    user_data['widget_broadcast_post_caption'] = message.html_text if message.text or message.caption else None
    if message.content_type != 'text':  # Upload non-text posts to Telegraph
        if message.photo:
            f = await (max(message.photo, key=lambda c: c.width)).download(destination=BytesIO())
        else:
            f = await getattr(message, message.content_type).download(destination=BytesIO())
        url = await telegraph_upload(f)
        user_data['widget_broadcast']['file_id'] = url
    user_data['widget_broadcast_content_type'] = message.content_type

    await neko.storage.set_user_data(data=user_data, user_id=message.from_user.id, replace=True)
    data = await neko.build_menu(name='widget_broadcast_post', obj=message)
    await data.send_message()


@ROUTER.function()
async def widget_broadcast_post_preview(_: Menu, call: Union[types.Message, types.CallbackQuery], neko: Neko):
    user_data = await neko.storage.get_user_data(user_id=call.from_user.id)

    await call.message.delete()

    await utils.send_post(user_data=user_data, chat_id=call.from_user.id, neko=neko)

    data = await neko.build_menu(name='widget_broadcast_post', obj=call)
    await data.send_message()


@ROUTER.function()
async def widget_broadcast_remove_button(data: Menu, call: Union[types.Message, types.CallbackQuery], neko: Neko):
    user_data = await neko.storage.get_user_data(user_id=call.from_user.id)
    row_index: int = int(data.call_data.split('-')[0])
    button_index: int = int(data.call_data.split('-')[1])
    post_markup: List[List[Dict[str, str]]] = user_data['widget_broadcast_post_markup']

    del post_markup[row_index][button_index]
    if len(post_markup[row_index]) == 0:
        del post_markup[row_index]
    user_data['widget_broadcast_post_markup'] = post_markup

    await neko.storage.set_user_data(data=user_data, user_id=call.from_user.id, replace=True)

    data = await neko.build_menu(name='widget_broadcast_post_markup', obj=call)

    await data.edit_text()


@ROUTER.function()
async def widget_broadcast_add_button_step_2(_: Menu, message: Union[types.Message, types.CallbackQuery], neko: Neko):
    if message.text.startswith('@'):
        message.text = message.text.replace('@', 'https://t.me/')
    user_data = await neko.storage.get_user_data(user_id=message.from_user.id)
    row_index: int = user_data['widget_broadcast_post_row_index']
    post_markup: List[List[Dict[str, str]]] = user_data.get('widget_broadcast_post_markup', [])
    if len(post_markup) == row_index:
        post_markup.append([])
    post_markup[row_index].append({'text': user_data['widget_broadcast_add_button_step_1']['text'],
                                   'url': message.text})
    user_data['widget_broadcast_post_markup'] = post_markup
    user_data.pop('widget_broadcast_post_row_index')
    user_data.pop('menu')
    await neko.storage.set_user_data(data=user_data, user_id=message.from_user.id, replace=True)
    data = await neko.build_menu(name='widget_broadcast_post_markup', obj=message)
    await data.send_message()


@ROUTER.function()
async def widget_broadcast_broadcast(data: Menu, call: Union[types.Message, types.CallbackQuery], neko: Neko):
    user_data = await neko.storage.get_user_data(user_id=call.from_user.id)

    total: int = await neko.storage.check('SELECT id FROM nekogram_users WHERE id != %s', call.from_user.id)
    attempts: int = 0
    successful: int = 0
    failed: int = 0

    await call.message.edit_text(text=data.text.format(total=total, attempts=0, successful=0, failed=0))

    async for user in neko.storage.select('SELECT id FROM nekogram_users WHERE id != %s', call.from_user.id):
        attempts += 1
        while True:
            try:
                await utils.send_post(user_data=user_data, chat_id=user['id'], neko=neko)
                successful += 1
                break
            except aiogram_exc.RetryAfter as e:
                await sleep(e.timeout)
            except aiogram_exc.TelegramAPIError:
                failed += 1
                break
            await sleep(.2)

        if attempts % 5 == 0:
            await call.message.edit_text(text=data.text.format(total=total, attempts=attempts,
                                                               successful=successful, failed=failed))

    for key in user_data.copy().keys():
        if key.startswith('widget_broadcast'):
            user_data.pop(key)
    user_data.pop('menu')
    await neko.storage.set_user_data(data=user_data, user_id=call.from_user.id, replace=True)
    await data.build(text_format={'total': total, 'attempts': attempts, 'successful': successful,
                                  'failed': failed}, allowed_buttons=[2])
    await data.edit_text()



from aiogram import types, exceptions as aiogram_exc
from typing import Union, Dict, Any
from contextlib import suppress
from ..logger import LOGGER
import NekoGram


async def default_start_handler(message: types.Message):
    neko: NekoGram.Neko = message.conf['neko']

    if not await neko.storage.check_user_exists(user_id=message.from_user.id):
        lang = message.from_user.language_code
        await neko.storage.create_user(user_id=message.from_user.id,
                                       language=lang if lang in neko.text_processor.texts.keys() else None)
    else:  # Reset user data on start
        await neko.storage.set_user_data(user_id=message.from_user.id)

    current_menu = await neko.build_menu(name='start', obj=message)
    if current_menu is None:
        raise RuntimeError(f'Start menu formatter should not break execution')

    if neko.functions.get('start'):
        await neko.functions['start'](current_menu, message, neko)
        return
    await current_menu.send_message()
    with suppress(Exception):
        await message.delete()


async def menu_message_handler(message: types.Message):
    neko: NekoGram.Neko = message.conf['neko']
    if message.text == '/start':  # Start case
        await default_start_handler(message)
        return

    user_data: Union[Dict[str, Any], bool] = await neko.storage.get_user_data(user_id=message.from_user.id)
    current_menu = await neko.build_menu(name=user_data['menu'], obj=message)  # Prebuild current menu
    if current_menu is None:
        return

    if message.text and message.text.startswith('⬅️'):  # Back button clicked
        await neko.storage.set_user_menu(menu=current_menu.prev_menu or None, user_id=message.from_user.id)
        last_message_id = await neko.storage.get_last_message_id(user_id=message.from_user.id)
        try:
            await neko.bot.delete_message(chat_id=message.from_user.id, message_id=last_message_id)
        except (aiogram_exc.MessageCantBeDeleted, aiogram_exc.MessageToDeleteNotFound):
            with suppress(Exception):
                await neko.bot.edit_message_reply_markup(chat_id=message.from_user.id, message_id=last_message_id)

        if neko.prev_menu_handlers.get(current_menu.name):
            current_menu.prev_menu = await neko.prev_menu_handlers[current_menu.name](current_menu)
        menu = await neko.build_menu(name=current_menu.prev_menu or 'start', obj=message)
        if menu is None:
            return
        await menu.send_message()

        with suppress(Exception):
            await message.delete()
        return

    if current_menu.filters:  # Check filters
        filters_passed: bool = False
        for f in current_menu.filters:
            if await neko.filters[f](message):
                filters_passed = True
                break
    else:
        filters_passed: bool = True

    if not filters_passed:  # Wrong data provided
        current_menu.text = current_menu.validation_error
        await current_menu.send_message()
        return
    else:
        await neko.storage.set_user_data(data={current_menu.name: message.to_python(),
                                               'menu': current_menu.next_menu}, user_id=message.from_user.id)

    if neko.functions.get(current_menu.name):  # Execute a function if present
        await neko.functions[current_menu.name](current_menu, message, neko)
        return

    if neko.next_menu_handlers.get(current_menu.name):
        current_menu.next_menu = await neko.next_menu_handlers[current_menu.name](current_menu)

    if not current_menu.next_menu:  # Next step is undefined
        LOGGER.warning(f'Unhandled user input for {current_menu.name}. *confused meow*')
    next_menu = await neko.build_menu(name=current_menu.next_menu or 'start', obj=message)
    if next_menu is None:
        return

    if neko.functions.get(next_menu.name) and not next_menu.filters:  # Execute a function in next menu if no filters
        await neko.functions[next_menu.name](next_menu, message, neko)

    await next_menu.send_message()



from aiogram.dispatcher.filters.builtin import ChatTypeFilter
from .text_processors import BaseProcessor, JSONProcessor
from .filters import StartsWith, HasMenu, BuiltInFilters
from aiogram import Bot, Dispatcher, executor, types
from .storages import BaseStorage, MemoryStorage
from typing import Optional, Dict, List, Union
from aiogram.dispatcher.filters import Filter
from abc import ABC, abstractmethod
from .logger import LOGGER
from . import handlers


class BaseNeko(ABC):
    def __init__(self, storage: Optional[BaseStorage] = None, token: Optional[str] = None, bot: Optional[Bot] = None,
                 dp: Optional[Dispatcher] = None, text_processor: Optional[BaseProcessor] = None,
                 entrypoint: str = 'start', menu_prefixes: Union[List[str]] = 'menu_',
                 callback_parameters_delimiter: str = '#'):
        """
        Initialize a Neko
        :param storage: A class that inherits from BaseDatabase class
        :param token: Telegram bot token
        :param bot: Aiogram Bot object
        :param dp: Aiogram Dispatcher object
        :param text_processor: A text processor to use (a class inherited from text_processors.BaseProcessor),
        JSONProcessor by default
        :param entrypoint: App entrypoint menu name (defaults to start)
        :param menu_prefixes: Common prefixes for menus defined in translation files
        :param callback_parameters_delimiter: A delimiter for callback data arguments
        """
        self.bot: Bot
        self.dp: Dispatcher
        if dp:
            self.bot, self.dp = (dp.bot, dp)
        elif bot:
            self.bot, self.dp = (bot, Dispatcher(bot=bot))
        elif token:
            self.bot = Bot(token=token)
            self.dp = Dispatcher(bot=self.bot)
        else:
            raise ValueError('No Dispatcher, Bot or token provided during Neko initialization')

        if storage is None:
            storage = MemoryStorage()
        self.storage: BaseStorage = storage

        if text_processor is None:
            text_processor = JSONProcessor()
        self.text_processor: BaseProcessor = text_processor

        if type(storage) == MemoryStorage:  # Check if MemoryStorage is used
            LOGGER.warning('You are using MemoryStorage which does not store data permanently! *sigh of condemnation*')

        if isinstance(menu_prefixes, str):
            menu_prefixes = [menu_prefixes]
        self._entrypoint: str = entrypoint
        self.menu_prefixes: List[str] = menu_prefixes
        self.filters: Dict[str, callable] = dict()
        self.callback_parameters_delimiter: str = callback_parameters_delimiter
        self.register_handlers()

        builtin_filters = BuiltInFilters()
        for f in builtin_filters.to_list:
            self.filters[f] = getattr(builtin_filters, f'is_{f}')

        print(r'''
   _  __    __        _____             
  / |/ /__ / /_____  / ___/______ ___ _ 
 /    / -_)  '_/ _ \/ (_ / __/ _ `/  ' \
/_/|_/\__/_/\_\\___/\___/_/  \_,_/_/_/_/''')

    def register_handlers(self):
        """
        Registers default handlers
        """
        self.dp.register_message_handler(handlers.menu_message_handler, ChatTypeFilter(types.ChatType.PRIVATE),
                                         commands=['start'])
        for menu_prefix in self.menu_prefixes:
            self.dp.register_callback_query_handler(handlers.menu_callback_query_handler, StartsWith(menu_prefix))
        if 'widget_' not in self.menu_prefixes:
            self.dp.register_callback_query_handler(handlers.menu_callback_query_handler, StartsWith('widget_'))
        self.dp.register_message_handler(handlers.menu_message_handler, ChatTypeFilter(types.ChatType.PRIVATE),
                                         HasMenu(self.storage), content_types=types.ContentType.ANY)

    def add_filter(self, name: str, callback: Union[callable, Filter]):
        if self.filters.get(name):
            LOGGER.warning(f'Filter named {name} was overridden. *eyes filling with tears*')

        if isinstance(callback, Filter):
            callback = callback.check
        self.filters[name] = callback

    def remove_filter(self, name: str):
        self.filters.pop(name)

    @abstractmethod
    async def check_text_exists(self, text: str, lang: Optional[str] = None) -> bool:
        pass

    def start_polling(self, on_startup: Optional[callable] = None, on_shutdown: Optional[callable] = None):
        executor.start_polling(self.dp, on_startup=on_startup, on_shutdown=on_shutdown)

    @abstractmethod
    def register_formatter(self, callback: callable, name: Optional[str] = None):
        pass

    @abstractmethod
    def formatter(self, name: Optional[str] = None):
        pass

    @abstractmethod
    def register_function(self, callback: callable, name: Optional[str] = None):
        pass

    @abstractmethod
    def function(self, name: Optional[str] = None):
        pass

    @abstractmethod
    def register_prev_menu_handler(self, callback: callable, name: Optional[str] = None):
        pass

    @abstractmethod
    def prev_menu_handler(self, name: Optional[str] = None):
        pass

    @abstractmethod
    def register_next_menu_handler(self, callback: callable, name: Optional[str] = None):
        pass

    @abstractmethod
    def next_menu_handler(self, name: Optional[str] = None):
        pass

    @abstractmethod
    def register_markup_overrider(self, callback: callable, lang: str, name: Optional[str] = None):
        pass

    @abstractmethod
    def markup_overrider(self, lang: str, name: Optional[str] = None):
        pass

    @abstractmethod
    async def build_menu(self, name: str, obj: Union[types.Message, types.CallbackQuery], user_id: Optional[int] = None,
                         callback_data: Optional[Union[str, int]] = None, auto_build: bool = True):
        pass



from typing import Optional, Union, Any, AsyncGenerator, List, Dict, Tuple

try:
    from psycopg2.extras import DictCursor
except ImportError:
    raise ImportError('Install aiopg to use PGStorage!')

from ..base_storage import BaseStorage
from aiopg.sa import create_engine
from psycopg2 import Error
import asyncio

try:
    import ujson as json
except ImportError:
    import json


class PGStorage(BaseStorage):
    def __init__(self, database: str, host: str = 'localhost', port: int = 5432, user: str = 'root',
                 password: Optional[str] = None, default_language: str = 'en'):
        """
        Initialize database
        :param database: Database name
        :param host: Database host
        :param port: Database port
        :param user: Database user
        :param password: Database password
        """

        self.engine = None
        self.host: str = host
        self.port: int = port
        self.user: str = user
        self.password: str = password
        self.database = database
        self.connection = None

        self.default_language = default_language

        asyncio.run(self.acquire_engine())
        super().__init__(default_language=default_language)

    def __del__(self):
        self.engine.terminate()
        await self.engine.wait_closed()

    async def acquire_engine(self):
        """
        Creates a new PostgreSQL engine
        """
        if self.engine:
            self.engine.terminate()
            await self.engine.wait_closed()

        self.engine = await create_engine(
            user=self.user,
            password=self.password,
            database=self.database,
            host=self.host,
            post=self.port
        )
        self.connection = await self.engine.acquire()

    @staticmethod
    def _verify_args(args: Optional[Union[Tuple[Union[Any, Dict[str, Any]], ...], Any]]):
        if args is None:
            args = tuple()
        if not isinstance(args, (tuple, dict)):
            args = (args,)
        return args

    async def apply(self, query: str, args: Optional[Union[Tuple[Any, ...], Dict[str, Any], Any]] = None):
        """
        Executes SQL query and returns the number of affected rows
        :param query: SQL query to execute
        :param args: Arguments passed to the SQL query
        :return: Number of affected rows
        """
        args = self._verify_args(args)
        try:
            result = await self.connection.execute(query, args)
            return await result.fetchone()
        except Error:
            return 0

    async def select(self, query: str, args: Optional[Union[Tuple[Any, ...], Dict[str, Any], Any]] = None) -> \
            AsyncGenerator[Dict[str, Any], None]:
        """
        Generator that yields rows
        :param query: SQL query to execute
        :param args: Arguments passed to the SQL query
        :return: Yields rows one by one
        """
        args = self._verify_args(args)
        try:
            async for row in self.connection.execute(query, args):
                yield row
        except Error:
            pass

    async def get(self, query: str, args: Optional[Union[Tuple[Any, ...], Dict[str, Any], Any]] = None,
                  fetch_all: bool = False) -> Union[bool, List[Dict[str, Any]], Dict[str, Any]]:
        """
        Get a single row or a list of rows from the database
        :param query: SQL query to execute
        :param args: Arguments passed to the SQL query
        :param fetch_all: Set True if you need a list of rows instead of just a single row
        :return: A row or a list or rows
        """
        args = self._verify_args(args)
        try:
            result = await self.connection.execute(query, args)
            if fetch_all:
                return await result.fetchall()
            else:
                return await result.fetchone()
        except Error:
            return False

    async def check(self, query: str, args: Optional[Union[Tuple[Any, ...], Dict[str, Any], Any]] = None) -> int:
        args = self._verify_args(args)
        try:
            result = await self.connection.execute(query, args)
            return result.rowcount
        except Error:
            return 0

    async def set_user_language(self, user_id: int, language: str):
        await super().set_user_language(user_id=user_id, language=language)
        await self.apply('UPDATE users SET lang = %s WHERE id = %s', (language, user_id))

    async def get_user_language(self, user_id: int) -> str:
        """
        Get user's language
        :param user_id: Telegram ID of the user
        :return: User's language
        """
        lang = await self.get_cached_user_language(user_id=user_id)
        if lang is None:
            lang = (await self.get('SELECT lang FROM users WHERE id=%s', user_id)).get('lang', self.default_language)
        await super().set_user_language(user_id=user_id, language=lang)
        return lang

    async def get_user_data(self, user_id: int) -> Union[Dict[str, Any], bool]:
        """
        Get user data
        :param user_id: Telegram ID of the user
        :return: Decoded JSON user data
        """
        try:
            return json.loads((await self.get('SELECT data FROM users WHERE id=%s', user_id)).get('data', '{}'))
        except TypeError:
            return False

    async def set_user_data(self, user_id: int, data: Optional[Dict[str, Any]] = None,
                            replace: bool = False) -> Dict[str, Any]:
        if data is None:
            data = dict()
            replace = True

        if replace:
            user_data = data
        else:
            user_data = await self.get_user_data(user_id=user_id)
            user_data.update(data)

        await self.apply('UPDATE users SET data = %s WHERE id = %s', (json.dumps(user_data), user_id))
        return user_data

    async def set_user_menu(self, user_id: int, menu: Optional[str] = None):
        await self.set_user_data(user_id=user_id, data={'menu': menu})
        return menu

    async def get_user_menu(self, user_id: int) -> Optional[str]:
        return (await self.get_user_data(user_id=user_id)).get('menu')

    async def check_user_exists(self, user_id: int) -> bool:
        return bool(await self.check('SELECT id FROM users WHERE id = %s', user_id))

    async def set_last_message_id(self, user_id: int, message_id: int):
        await self.apply('UPDATE users SET last_message_id = %s WHERE id = %s', (message_id, user_id))

    async def get_last_message_id(self, user_id: int) -> Optional[int]:
        return (await self.get('SELECT last_message_id FROM users WHERE id = %s', user_id)).get('last_message_id')

    async def create_user(self, user_id: int, language: Optional[str] = None):
        if language is None:
            language = self.default_language

        await self.apply('INSERT INTO users (id, lang) VALUES (%s, %s)', (user_id, language))



from typing import Union, Dict, Any, TextIO
from abc import ABC, abstractmethod


class BaseProcessor(ABC):
    def __init__(self):
        self.texts: Dict[str, Dict[str, Any]] = dict()

    @staticmethod
    @abstractmethod
    def add_texts(texts: Union[Dict[str, Any], TextIO, str] = 'translations', is_widget: bool = False):
        pass

    @property
    @abstractmethod
    def processor_type(self) -> str:
        return 'base'



from typing import Dict, List, Union, Optional, Any, Callable, Awaitable
from .text_processors import BaseProcessor
from aiogram import Dispatcher, Bot, types
from .storages.mysql import MySQLStorage
from .utils import HandlerInjector
from .storages import BaseStorage
from .base_neko import BaseNeko
from .router import NekoRouter
from datetime import datetime
from .logger import LOGGER
import importlib_resources
from copy import deepcopy
from .menus import Menu

try:
    import ujson as json
except ImportError:
    import json


class Neko(BaseNeko):
    _registration_warned: bool = False
    _builtin_widgets: List[str] = ['broadcast', 'languages']
    _widgets_warned: bool = False

    def __init__(self, storage: Optional[BaseStorage] = None, token: Optional[str] = None, bot: Optional[Bot] = None,
                 dp: Optional[Dispatcher] = None, text_processor: Optional[BaseProcessor] = None,
                 menu_prefixes: Union[List[str]] = 'menu_', load_texts: bool = True,
                 callback_parameters_delimiter: str = '#'):
        super().__init__(storage=storage, token=token, bot=bot, dp=dp, text_processor=text_processor,
                         menu_prefixes=menu_prefixes, callback_parameters_delimiter=callback_parameters_delimiter)
        self.dp.middleware.setup(HandlerInjector(self))  # Set up the handler injector middleware

        self._cached_user_languages: Dict[str, Dict[str, Union[str, datetime]]] = dict()
        if load_texts:
            self.text_processor.add_texts()
        self.functions: Dict[str, Callable[[Menu, Union[types.Message, types.CallbackQuery], BaseNeko],
                                           Awaitable[Any]]] = dict()
        self.format_functions: Dict[str, Callable[[Menu, types.User, BaseNeko], Awaitable[Any]]] = dict()
        self.prev_menu_handlers: Dict[str, Callable[[Menu], Awaitable[str]]] = dict()
        self.next_menu_handlers: Dict[str, Callable[[Menu], Awaitable[str]]] = dict()
        self._markup_overriders: Dict[str, Dict[str, Callable[[Menu], Awaitable[List[List[Dict[str, str]]]]]]] = dict()
        self.widgets: List[str] = list()

    async def check_text_exists(self, text: str, lang: Optional[str] = None) -> bool:
        if lang is None:
            lang = list(self.text_processor.texts.keys())[0]
        return text in self.text_processor.texts[lang].keys()

    def register_formatter(self, callback: Callable[[Menu, types.User, BaseNeko], Any], name: Optional[str] = None):
        """
        Register a formatter
        :param callback: A formatter to call
        :param name: Menu name
        """
        if not self._registration_warned:
            LOGGER.warning('It is not recommended to register formatters within a Neko class, '
                           'consider using a NekoRouter')
            self._registration_warned = True
        self.format_functions[name or callback.__name__] = callback

    def formatter(self, name: Optional[str] = None):
        """
        Register a formatter
        :param name: Menu name
        """

        def decorator(callback: Callable[[Menu, types.User, BaseNeko], Any]):
            self.register_formatter(callback=callback, name=name)
            return callback

        return decorator

    def register_function(self, callback: Callable[[Menu, Union[types.Message, types.CallbackQuery], BaseNeko], Any],
                          name: Optional[str] = None):
        """
        Register a function
        :param callback: A function to call
        :param name: Menu name
        """
        if not self._registration_warned:
            LOGGER.warning('It is not recommended to register functions within a Neko class, '
                           'consider using a NekoRouter')
            self._registration_warned = True
        self.functions[name or callback.__name__] = callback

    def function(self, name: Optional[str] = None):
        """
        Register a function
        :param name: Menu name
        """

        def decorator(callback: Callable[[Menu, Union[types.Message, types.CallbackQuery], BaseNeko], Any]):
            self.register_function(callback=callback, name=name)
            return callback

        return decorator

    def register_prev_menu_handler(self, callback: Callable[[Menu], Awaitable[str]], name: Optional[str] = None):
        """
        Register a prev menu handler
        :param callback: A prev menu handler to call
        :param name: Menu name
        """
        self.prev_menu_handlers[name or callback.__name__] = callback

    def prev_menu_handler(self, name: Optional[str] = None):
        """
        Register a prev menu handler
        :param name: Menu name
        """

        def decorator(callback: Callable[[Menu], Awaitable[str]]):
            self.register_prev_menu_handler(callback=callback, name=name)
            return callback

        return decorator

    def register_next_menu_handler(self, callback: Callable[[Menu], Awaitable[str]], name: Optional[str] = None):
        """
        Register a next menu handler
        :param callback: A next menu handler to call
        :param name: Menu name
        """
        self.prev_menu_handlers[name or callback.__name__] = callback

    def next_menu_handler(self, name: Optional[str] = None):
        """
        Register a next menu handler
        :param name: Menu name
        """

        def decorator(callback: Callable[[Menu], Awaitable[str]]):
            self.register_next_menu_handler(callback=callback, name=name)
            return callback

        return decorator

    def register_markup_overrider(self, callback: Callable[[Menu], Awaitable[List[List[Dict[str, str]]]]], lang: str,
                                  name: Optional[str] = None):
        """
        Register a markup overrider
        :param callback: A markup overrider to call
        :param lang: Language to override markup for
        :param name: Menu name
        """
        if name is None:
            name = callback.__name__
        if name not in self._markup_overriders:
            self._markup_overriders[name] = dict()
        self._markup_overriders[name][lang] = callback

    def markup_overrider(self, lang: str, name: Optional[str] = None):
        """
        Register a markup overrider
        :param lang: Language to override markup for
        :param name: Menu name
        """

        def decorator(callback: Callable[[Menu], Awaitable[List[List[Dict[str, str]]]]]):
            self.register_markup_overrider(callback=callback, name=name, lang=lang)
            return callback

        return decorator

    async def build_menu(self, name: str, obj: Union[types.Message, types.CallbackQuery],
                         user_id: Optional[int] = None,
                         callback_data: Optional[Union[str, int]] = None,
                         auto_build: bool = True) -> Optional[Menu]:
        """
        Build a menu by its name
        :param name: Menu name, same as in translation file
        :param obj: An Aiogram Message or CallbackQuery object
        :param user_id: An ID of a user to build menu for
        :param callback_data: Callback data to assign to a menu
        :param auto_build: Whether to build the Menu if no formatter is defined
        :return: A Menu object
        """
        if name == 'menu_start':  # Start patch
            name = 'start'

        lang = await self.storage.get_user_language(user_id=user_id or obj.from_user.id)
        text = deepcopy(self.text_processor.texts[lang].get(name))
        if not obj:
            raise RuntimeError(f'Neither Message nor CallbackQuery was provided during menu building for {name}! '
                               f'*brain explosion sounds accompanied by intense meowing*')
        if text is None:  # Try to fetch the menu in another language
            if 'en' in self.text_processor.texts.keys():
                text = self.text_processor.texts['en'].get(name)
            if text is None:
                for language in self.text_processor.texts.keys():
                    text = self.text_processor.texts[language].get(name)
                    if text:
                        LOGGER.warning(f'{name} menu does not have {lang} translation, using {language}.')
                        break
            else:
                LOGGER.warning(f'{name} menu does not have {lang} translation, using en.')
            if text is None:
                raise RuntimeError(f'There is no menu called {name}! *facePAWm*')
        if text.get('text') is None and text:
            LOGGER.warning(f'No text provided for {name}. *suspicious stare*')
        menu = Menu(name=name, obj=obj, markup=text.get('markup'), markup_row_width=text.get('markup_row_width'),
                    text=text.get('text'), no_preview=text.get('no_preview'), parse_mode=text.get('parse_mode'),
                    silent=text.get('silent'), validation_error=text.get('validation_error'),
                    extras=text.get('extras'), keyboard_values_to_format=text.get('keyboard_values_to_format'),
                    markup_type=text.get('markup_type'), prev_menu=text.get('prev_menu'),
                    next_menu=text.get('next_menu'), filters=text.get('filters'), callback_data=callback_data)

        if self._markup_overriders.get(name, dict()).get(lang):
            menu.raw_markup = await self._markup_overriders[name][lang](menu)

        format_func = self.format_functions.get(name)
        if format_func:
            await format_func(menu, obj.from_user, self)
            if menu.markup is None and menu.raw_markup:
                await menu.build()
        elif auto_build:
            await menu.build()

        return menu if not menu.is_broken else None

    def attach_router(self, router: NekoRouter):
        """
        Attach a NekoRouter to Neko
        :param router: A NekoRouter to attach
        """
        router.attach()
        self.functions.update(router.functions)
        self.format_functions.update(router.format_functions)
        self.prev_menu_handlers.update(router.prev_menu_handlers)
        self.next_menu_handlers.update(router.next_menu_handlers)

    async def attach_widget(self, formatters_router: NekoRouter, functions_router: NekoRouter,
                            startup: Callable[[BaseNeko], Awaitable[Any]], texts_path: Optional[str] = None,
                            db_table_structure_path: Optional[str] = None,
                            formatters_to_ignore: Optional[List[str]] = None,
                            functions_to_ignore: Optional[List[str]] = None):
        """
        Attach a widget to Neko
        :param formatters_router: A NekoRouter object responsible for formatters
        :param functions_router: A NekoRouter object responsible for functions
        :param startup: A startup function to call
        :param texts_path: A path to translation files
        :param db_table_structure_path: A path to table structure file
        :param formatters_to_ignore: A list of formatter names to ignore
        :param functions_to_ignore: A list of function names to ignore
        """
        if not isinstance(self.storage, MySQLStorage) and not self._widgets_warned:
            LOGGER.warning(f'Your storage is not MySQLStorage, widgets may function improperly.')
            self._widgets_warned = True
        if formatters_router.name is None or formatters_router.name != functions_router.name \
                or functions_router.name is None:
            raise RuntimeError('Widget router names must be present and must be same for formatter and function router')

        if formatters_router.name in self.widgets:
            LOGGER.warning(f'Widget {formatters_router.name} is being attached again, ignored. *ultrasonic meowing*')
            return

        await startup(self)
        self.widgets.append(formatters_router.name)

        if formatters_to_ignore:  # Ignore formatters
            for name, func in formatters_router.format_functions.copy().items():
                if name in formatters_to_ignore:
                    formatters_router.format_functions.pop(name)

        if functions_to_ignore:  # Ignore functions
            for name, func in functions_router.functions.copy().items():
                if name in functions_to_ignore:
                    functions_router.functions.pop(name)

        self.attach_router(formatters_router)
        self.attach_router(functions_router)

        if db_table_structure_path and isinstance(self.storage, MySQLStorage):
            with open(db_table_structure_path, 'r') as f:
                table_structure = json.load(f)
            await self.storage.add_tables(table_structure, required_by=formatters_router.name)

        if texts_path is None:
            if formatters_router.name in self._builtin_widgets:
                for x in importlib_resources.files(f'NekoGram.widgets.{formatters_router.name}.translations').iterdir():
                    print(str(x), x)
                    self.text_processor.add_texts(str(x), is_widget=True)
            else:
                raise RuntimeError(f'Widget {formatters_router.name} is not builtin, '
                                   f'therefore texts_path has to be provided')
        LOGGER.info(f'{formatters_router.name.capitalize()} widget attached successfully')



from typing import Union, Optional, Dict, Any, AsyncGenerator, List
from .base_storage import BaseStorage
from ..logger import LOGGER


class MemoryStorage(BaseStorage):
    def __init__(self, default_language: str = 'en'):
        super().__init__(default_language=default_language)
        self._message_history: Dict[str, int] = dict()

    async def set_user_language(self, user_id: int, language: str):
        await super().set_user_language(user_id=user_id, language=language)
        if not self.users.get(str(user_id)):  # Check if user exists so it won't fail
            self.users[str(user_id)] = {}

        self.users[str(user_id)]['language'] = language
        LOGGER.info(f'Set language to {language} for {user_id}')

    async def get_user_language(self, user_id: int) -> str:
        lang = self.users.get(str(user_id), {}).get('language', self.default_language)
        await super().set_user_language(user_id=user_id, language=lang)
        LOGGER.info(f'Retrieved language for {user_id}: {lang}')
        return lang

    async def set_user_data(self, user_id: int, data: Optional[Dict[str, Any]] = None,
                            replace: bool = False) -> Dict[str, Any]:
        if data is None:
            data = dict()
            replace = True

        if replace:
            user_data = data
        else:
            user_data = await self.get_user_data(user_id=user_id)
            user_data.update(data)

        if not self.users.get(str(user_id)):  # Check if user exists so it won't fail
            self.users[str(user_id)] = {}

        self.users[str(user_id)]['data'] = user_data
        LOGGER.info(f'Set user data for {user_id}: {user_data}')
        return user_data

    async def get_user_data(self, user_id: int) -> Dict[str, Any]:
        data = self.users.get(str(user_id), {}).get('data', {})
        LOGGER.info(f'Fetched user data for {user_id}: {data}')
        return data

    async def set_user_menu(self, user_id: int, menu: Optional[str] = None):
        await self.set_user_data(user_id=user_id, data={'menu': menu})
        return menu

    async def get_user_menu(self, user_id: int) -> Optional[str]:
        return (await self.get_user_data(user_id=user_id)).get('menu')

    async def check_user_exists(self, user_id: int) -> bool:
        exists: bool = bool(self.users.get(str(user_id)))
        LOGGER.info(f'Checked existence of {user_id}: {exists}')
        return exists

    async def set_last_message_id(self, user_id: int, message_id: int):
        LOGGER.info(f'Set last message ID for {user_id}: {message_id}')
        self._message_history[str(user_id)] = message_id

    async def get_last_message_id(self, user_id: int) -> Optional[int]:
        message_id = self._message_history.get(str(user_id))
        LOGGER.info(f'Retrieved last message ID for {user_id}: {message_id}')
        return message_id

    async def create_user(self, user_id: int, language: Optional[str] = None):
        if language is None:
            language = self.default_language

        self.users[str(user_id)] = {'language': language}
        LOGGER.info(f'Created a new user: {user_id}, {language}')

    async def apply(self, query, args=None):
        pass

    async def select(self, query, args=None) -> AsyncGenerator[Dict[str, Any], None]:
        yield

    async def get(self, query: str, args=None, fetch_all: bool = False) -> Union[bool, List[Dict[str, Any]],
                                                                                 Dict[str, Any]]:
        pass

    async def check(self, query: str, args=None) -> int:
        pass



from aiogram.dispatcher.filters import Filter
from aiogram.types import Message, CallbackQuery
from typing import Dict, List, Any, Union, Optional
from .storages.base_storage import BaseStorage
import re


class HasMenu(Filter):
    def __init__(self, database: BaseStorage):
        self.database: BaseStorage = database

    @classmethod
    def validate(cls, _: Dict[str, Any]):
        return {}

    async def check(self, obj: Union[Message, CallbackQuery]) -> bool:
        return bool((await self.database.get_user_data(user_id=obj.from_user.id)).get('menu', False))


class StartsWith(Filter):
    """
    Filter for checking if message text or callback query starts with a certain text
    """

    def __init__(self, starts_with: Optional[Union[List[str], str]] = None):
        if not starts_with:
            raise ValueError('Text to start with is not specified!')

        self.starts_with: List[str] = starts_with if isinstance(starts_with, list) else [starts_with]

    @classmethod
    def validate(cls, _: Dict[str, Any]):
        return {}

    async def check(self, obj: Union[Message, CallbackQuery]) -> bool:
        for text in self.starts_with:
            if isinstance(obj, Message) and obj.text.startswith(text):
                return True
            elif isinstance(obj, CallbackQuery) and obj.data.startswith(text):
                return True
            return False

        return False


class BuiltInFilters:
    @staticmethod
    async def _to_message(obj: Union[Message, CallbackQuery]) -> Message:
        if isinstance(obj, CallbackQuery):
            obj = obj.message
        return obj

    @staticmethod
    async def is_any(_: Union[Message, CallbackQuery]) -> bool:
        """
        Check if message is of any content types available
        :return: Always True
        """
        return True

    async def is_int(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message text can be converted to an integer
        :return: True if so
        """
        obj = await self._to_message(obj)
        return obj.text and obj.text.isdigit()

    @staticmethod
    async def is_int_neg(obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message text can be converted to a negative integer
        :return: True if so
        """
        try:
            return int(obj.text) < 0
        except ValueError:
            return False

    @staticmethod
    async def is_int_pos(obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message text can be converted to a positive integer
        :return: True if so
        """
        try:
            return int(obj.text) > 0
        except ValueError:
            return False

    async def is_float(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message text can be converted to a float
        :return: True if so
        """
        obj = await self._to_message(obj)
        return obj.text and obj.text.isnumeric()

    async def is_text(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message content is text
        :return: True if so
        """
        obj = await self._to_message(obj)
        return bool(obj.text)

    async def is_photo(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message content is a photo
        :return: True if so
        """
        obj = await self._to_message(obj)
        return obj.content_type == 'photo'

    async def is_video(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message content is a video
        :return: True if so
        """
        obj = await self._to_message(obj)
        return obj.content_type == 'video'

    async def is_animation(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message content is a GIF
        :return: True if so
        """
        obj = await self._to_message(obj)
        return obj.content_type == 'animation'

    async def is_http_url(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message text is an HTTP URL
        :return: True if so
        """
        obj = await self._to_message(obj)
        return obj.text and re.fullmatch(r'http://[-a-zA-Z0-9@:%._+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b(?:'
                                         r'[-a-zA-Z0-9()@:%_+.~#?&/=]*)$', obj.text)

    async def is_https_url(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message text is an HTTPS URL
        :return: True if so
        """
        obj = await self._to_message(obj)
        return obj.text and re.fullmatch(r'https://[-a-zA-Z0-9@:%._+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b(?:'
                                         r'[-a-zA-Z0-9()@:%_+.~#?&/=]*)$', obj.text)

    async def is_tg_url(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message text is a Telegram URL
        :return: True if so
        """
        obj = await self._to_message(obj)
        return obj.text and re.fullmatch(r'tg://[-a-zA-Z0-9_?=]+', obj.text)

    async def is_mention(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message text is a Telegram user mention
        :return: True if so
        """
        obj = await self._to_message(obj)
        return obj.text and re.fullmatch(r'@[a-zA-Z0-9_]+', obj.text)

    async def is_url(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message text is an HTTP/HTTPS/Telegram URL
        :return: True if so
        """
        obj = await self._to_message(obj)
        return obj.text and re.fullmatch(r'(http|https|tg)://[-a-zA-Z0-9@:%._+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b(?:'
                                         r'[-a-zA-Z0-9()@:%_+.~#?&/=]*)$', obj.text)

    async def is_email(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message text is an email
        :return: True if so
        """
        obj = await self._to_message(obj)
        return obj.text and re.fullmatch(r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b', obj.text)

    async def is_phone_number(self, obj: Union[Message, CallbackQuery]) -> bool:
        """
        Checks if message text is an international phone number
        :return: True if so
        """
        obj = await self._to_message(obj)
        return obj.text and re.fullmatch(r'\+([0-9]+\s*)?([0-9]+)?[\s0-9\-]+[0-9]+', obj.text)

    @property
    def to_list(self):
        return [method.replace('is_', '') for method in dir(self) if not method.startswith(('_', 'to_list'))]



from typing import Callable, Optional, Union, Dict, List, Any, Awaitable
from .base_neko import BaseNeko
from aiogram import types
from .menus import Menu


class NekoRouter:
    def __init__(self, name: Optional[str] = None):
        self.functions: Dict[str, Callable[[Menu, Union[types.Message, types.CallbackQuery], BaseNeko],
                                           Awaitable[Any]]] = dict()
        self.format_functions: Dict[str, Callable[[Menu, types.User, BaseNeko], Awaitable[Any]]] = dict()
        self.prev_menu_handlers: Dict[str, Callable[[Menu], Awaitable[str]]] = dict()
        self.next_menu_handlers: Dict[str, Callable[[Menu], Awaitable[str]]] = dict()
        self.name = name
        self._was_attached: bool = False

    def register_formatter(self, callback: Callable[[Menu, types.User, BaseNeko], Any], name: Optional[str] = None):
        """
        Register a formatter
        :param callback: A formatter to call
        :param name: Menu name
        """
        self.format_functions[name or callback.__name__] = callback

    def formatter(self, name: Optional[str] = None):
        """
        Register a formatter
        :param name: Menu name
        """

        def decorator(callback: Callable[[Menu, types.User, BaseNeko], Any]):
            self.register_formatter(callback=callback, name=name)
            return callback

        return decorator

    def register_function(self, callback: Callable[[Menu, Union[types.Message, types.CallbackQuery], BaseNeko], Any],
                          name: Optional[str] = None):
        """
        Register a function
        :param callback: A function to call
        :param name: Menu name
        """
        self.functions[name or callback.__name__] = callback

    def function(self, name: Optional[str] = None):
        """
        Register a function
        :param name: Menu name
        """

        def decorator(callback: Callable[[Menu, Union[types.Message, types.CallbackQuery], BaseNeko], Any]):
            self.register_function(callback=callback, name=name)
            return callback

        return decorator

    def attach(self):
        if self._was_attached:
            raise RuntimeError('You are trying to attach the same router twice! *mad meowing*')
        self._was_attached = True

    def register_prev_menu_handler(self, callback: Callable[[Menu], Awaitable[str]], name: Optional[str] = None):
        """
        Register a prev menu handler
        :param callback: A prev menu handler to call
        :param name: Menu name
        """
        self.prev_menu_handlers[name or callback.__name__] = callback

    def prev_menu_handler(self, name: Optional[str] = None):
        """
        Register a prev menu handler
        :param name: Menu name
        """

        def decorator(callback: Callable[[Menu], Awaitable[str]]):
            self.register_prev_menu_handler(callback=callback, name=name)
            return callback

        return decorator

    def register_next_menu_handler(self, callback: Callable[[Menu], Awaitable[str]], name: Optional[str] = None):
        """
        Register a next menu handler
        :param callback: A next menu handler to call
        :param name: Menu name
        """
        self.prev_menu_handlers[name or callback.__name__] = callback

    def next_menu_handler(self, name: Optional[str] = None):
        """
        Register a next menu handler
        :param name: Menu name
        """

        def decorator(callback: Callable[[Menu], Awaitable[str]]):
            self.register_next_menu_handler(callback=callback, name=name)
            return callback

        return decorator
